#ifndef _CHISE_H

#include <db.h>
#include <errno.h>


typedef enum CHISE_DS_Type
{
  CHISE_DS_NONE,
  CHISE_DS_Berkeley_DB
} CHISE_DS_Type;

typedef struct CHISE_DS
{
  CHISE_DS_Type type;
  char *location;
} CHISE_DS;

int chise_open_data_source (CHISE_DS *ds, CHISE_DS_Type type,
			    char *location);

int chise_close_data_source (CHISE_DS *ds);


typedef int CHISE_Char_ID;


typedef DBT CHISE_Value;

static inline int
chise_value_size (const CHISE_Value *s)
{
  return s->size;
}

static inline char *
chise_value_to_c_string (const CHISE_Value *s)
{
  return s->data;
}


typedef DB CHISE_Attribute_Table;


typedef CHISE_Attribute_Table CHISE_Decoding_Table;

int chise_open_decoding_table (CHISE_Decoding_Table **db,
			       CHISE_DS *ds, const char *ccs,
			       DBTYPE real_subtype,
			       u_int32_t accessmask, int modemask);

int chise_close_decoding_table (CHISE_Decoding_Table *db);

CHISE_Char_ID chise_dt_get_char (CHISE_Decoding_Table *db, int code_point);


typedef CHISE_Attribute_Table CHISE_Feature_Table;

int chise_open_feature_table (CHISE_Feature_Table **db,
			      CHISE_DS *ds, const char *feature,
			      DBTYPE real_subtype,
			      u_int32_t accessmask, int modemask);

int chise_close_feature_table (CHISE_Feature_Table *db);

int chise_ft_get_value (CHISE_Feature_Table *db,
			CHISE_Char_ID cid, CHISE_Value *valdatum);



CHISE_Char_ID chise_decode_char (CHISE_DS *ds, char *ccs, int code_point);

int chise_get_feature (CHISE_DS *ds, CHISE_Char_ID cid,
		       char *key, CHISE_Value *valdatum);


int chise_open_attribute_table (CHISE_Attribute_Table **db,
				const char *db_dir,
				const char *encoding, const char *feature,
				DBTYPE real_subtype,
				u_int32_t accessmask, int modemask);

int chise_close_attribute_table (CHISE_Attribute_Table *db);

int chise_get_attribute_table (CHISE_Attribute_Table *db,
			       char *key, CHISE_Value *valdatum);

#endif /* !_CHISE_H */
